/***************************************************************************
                          titlewidget  -  description
                             -------------------
    begin                : Mon Oct 31 2005
    copyright            : (C) 2005 by Jason Wood
    email                : jasonwood@blueyonder.co.uk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qlayout.h>
#include <qfont.h>
#include <qpoint.h>
#include <qcolor.h>
#include <qwmatrix.h>
#include <qspinbox.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qbitmap.h>
#include <qimage.h>
#include <qtoolbutton.h>
#include <qcursor.h>

#include <kpushbutton.h>
#include <kfontcombo.h>
#include <kcolorcombo.h>
#include <kinputdialog.h>
#include <kdebug.h>
#include <kiconloader.h>
#include <ktempfile.h>

#include "titlewidget.h"

#define CursorMode 1
#define TextMode 2
#define RectMode 3
#define ResizeMode 4

//TODO don't hardcode image size
#define imageWidth 720
#define imageHeight 576
// safety margin for text
#define horizontalMarginSize 20
#define verticalMarginSize 20

FigureEditor::FigureEditor(
        QCanvas& c, QWidget* parent,
        const char* name, WFlags f) :
                QCanvasView(&c,parent,name,f)
{
        //Create temp file that will be used for preview in the Mlt monitor
        tmp=new KTempFile(QString::null,".png");

        selection=0;
        numItems=0;
        drawingRect=0;
        selectedItem=0;

        // Enable focus to grab keyboard events
        setFocusPolicy(QWidget::StrongFocus);
        setFocus();

        //TODO make background color configurable
        canvas()->setBackgroundColor(black);

        // Draw rectangle showing safety margins for the text
        QCanvasRectangle *marginRect = new QCanvasRectangle(QRect(horizontalMarginSize,verticalMarginSize,imageWidth-(2*horizontalMarginSize),imageHeight-(2*verticalMarginSize)),canvas());
        marginRect->setZ(-100);
        marginRect->setPen(QPen(QColor(255,255,255)));
        marginRect->show();
}


FigureEditor::~FigureEditor()
{
        // Delete temp file when it is not needed anymore
        tmp->unlink();
}


void FigureEditor::resizeEvent ( QResizeEvent * e)
{
        //TODO make canvas keep a fixed ratio when resizing
        QWMatrix wm;
        wm.scale(((double) width()-10)/((double) imageWidth),((double) height()-10)/((double) imageHeight));
        setWorldMatrix (wm);
}


void FigureEditor::contentsMouseDoubleClickEvent(QMouseEvent* e)
{
        // Double clicking on a text item opens the text edit widget
        if (selection)
                delete selection;
        selection=0;
        QPoint p = inverseWorldMatrix().map(e->pos());
        QCanvasItemList l=canvas()->collisions(p);
        if (l.isEmpty())
                return;
        if ( (l.first())->rtti() == 3 )
                emit editCanvasItem((QCanvasText*)(l.first()));
}


void FigureEditor::contentsMouseReleaseEvent(QMouseEvent* e)
{
        // If user was resizing replace rect with the new one
        if (operationMode == ResizeMode) {
                int pos=selectedItem->z();
                delete selectedItem;
                delete drawingRect;
                drawingRect=0;
                QPoint p = inverseWorldMatrix().map(e->pos());
                emit addRect(QRect(draw_start,p),pos);
                return;
        }
        // If user was drawing a new rectangle, create it
        if (operationMode == RectMode) {
                delete drawingRect;
                drawingRect=0;
                QPoint p = inverseWorldMatrix().map(e->pos());
                if ((p-draw_start).manhattanLength()>20)
                        emit addRect(QRect(draw_start,p),-1);
                // If new rectangle is too tiny, don't create it, was probably a user mistake
                else {
                        operationMode=CursorMode;
                        setCursor(arrowCursor);
                        canvas()->update();
                        emit adjustButtons();
                }
        }
        // If user was moving an item, end the move
        else if (moving) {
                moving=0;
                setCursor(QCursor(Qt::ArrowCursor));
        }
}

void FigureEditor::keyPressEvent ( QKeyEvent * e )
{
        // delete item on del key press
        if (e->key()==Qt::Key_Delete) {
                if (selectedItem) {
                        delete selectedItem;
                        delete selection;
                        selection=0;
                        canvas()->update();
                }
        } else
                e->ignore();
}


void FigureEditor::itemUp()
{
        // move item up
        if (!selectedItem)
                return;
        QCanvasItemList list = canvas()->allItems();
        QCanvasItemList::Iterator it = list.begin();
        for (; it != list.end(); ++it)
                if ((*it)->z() == selectedItem->z()+1) {
                        (*it)->setZ(selectedItem->z());
                        selectedItem->setZ(selectedItem->z()+1);
                        break;
                }
        canvas()->update();
}


void FigureEditor::itemDown()
{
        // move item down
        if (!selectedItem)
                return;
        QCanvasItemList list = canvas()->allItems();
        QCanvasItemList::Iterator it = list.begin();
        for (; it != list.end(); ++it)
                if ((*it)->z() == selectedItem->z()-1) {
                        (*it)->setZ(selectedItem->z());
                        selectedItem->setZ(selectedItem->z()-1);
                        break;
                }
        canvas()->update();
}


void FigureEditor::deleteItem(QCanvasItem *i)
{
        // delete item
        QCanvasItemList list = canvas()->allItems();
        QCanvasItemList::Iterator it = list.begin();
        for (; it != list.end(); ++it)
                if ((*it)->z()<=i->z())
                        (*it)->setZ((*it)->z()-1);
        delete i;
}


void FigureEditor::startResize(QPoint p)
{
        // User wants to resize a rectangle
        delete selection;
        selection=0;
        operationMode=ResizeMode;
        draw_start=p;
        canvas()->update();
}

void FigureEditor::contentsMousePressEvent(QMouseEvent* e)
{
        QPoint p = inverseWorldMatrix().map(e->pos());

        // Create new item if user wants to
        if (operationMode!=CursorMode) {
                delete selection;
                selection=0;
                moving = 0;
                selectedItem = 0;
                if (operationMode == TextMode)
                        emit addText(p);
                if (operationMode == RectMode)
                        draw_start=p;
                return;
        }

        // Deselect item if user clicks in an empty area
        QCanvasItemList l=canvas()->collisions(p);
        if (l.isEmpty() || (l.first()->z()<0)) {
                if (selection)
                        delete selection;
                selection = 0;
                moving = 0;
                selectedItem=0;
                canvas()->update();
                return;
        }

	QCanvasItemList::Iterator it=l.begin();
        // If user clicked in a rectangle corner, start resizing
        if (*it == selection) {
                uint dist=(p-(*selection).rect().topLeft()).manhattanLength();
                if (dist<20) {
                        startResize((*selection).rect().bottomRight());
                        return;
                }
                dist=(p-(*selection).rect().topRight()).manhattanLength();
                if (dist<20) {
                        startResize((*selection).rect().bottomLeft());
                        return;
                }
                dist=(p-(*selection).rect().bottomRight()).manhattanLength();
                if (dist<20) {
                        startResize((*selection).rect().topLeft());
                        return;
                }
                dist=(p-(*selection).rect().bottomLeft()).manhattanLength();
                if (dist<20) {
                        startResize((*selection).rect().topRight());
                        return;
                }
                it++;
        }

        // Otherwise, select item and prepare for moving
        moving = *it;
        selectedItem = moving;
        if (selection)
                delete selection;
        selection=0;

        if ( (*it)->rtti() == 3 ) {
                selectRectangle(*it);
                emit selectedCanvasItem((QCanvasText*)(*it));
        }
        if ( (*it)->rtti() == 5 ) {
                selectRectangle(*it);
                emit selectedCanvasItem((QCanvasRectangle*)(*it));
        }
        moving_start = p;
        canvas()->update();
}


void FigureEditor::selectRectangle(QCanvasItem *it)
{
        // Draw selection rectangle around selected item
        if ( (it)->rtti() == 3 )
                selection = new QCanvasRectangle(((QCanvasText*)(it))->boundingRect (),canvas());
        else
                if ( (it)->rtti() == 5 )
                        selection = new QCanvasRectangle(((QCanvasRectangle *)(it))->rect(),canvas());
        // set its Z index to 1000, so it is not drawn by the export routine
        selection->setZ(1000);
	QPen pen = QPen(QColor(120,60,60));
        pen.setStyle(Qt::DotLine);
        selection->setPen(pen);
        selection->show();
        canvas()->update();
}

void FigureEditor::clear()
{
        // clear all canvas
        QCanvasItemList list = canvas()->allItems();
        QCanvasItemList::Iterator it = list.begin();
        for (; it != list.end(); ++it) {
                if ( *it )
                        delete *it;
        }
}


void FigureEditor::changeTextSize(int newSize)
{
        if (!selectedItem || selectedItem->rtti()!=3)
                return;
        ((QCanvasText*)(selectedItem))->setFont(QFont(((QCanvasText*)(selectedItem))->font().family(),newSize));

        // Update selection rectangle
        delete selection;
        selection = new QCanvasRectangle(selectedItem->boundingRect (),canvas());
        selection->setZ(1000);
	QPen pen = QPen(QColor(120,60,60));
        pen.setStyle(Qt::DotLine);
        selection->setPen(pen);
        selection->show();
        canvas()->update();
}

void FigureEditor::changeTextFace(const QString & newFace)
{
        if (!selectedItem || selectedItem->rtti()!=3)
                return;
        ((QCanvasText*)(selectedItem))->setFont(QFont(newFace,((QCanvasText*)(selectedItem))->font().pointSize()));

        // Update selection rectangle
        delete selection;
        selection = new QCanvasRectangle(selectedItem->boundingRect (),canvas());
        selection->setZ(1000);
	QPen pen = QPen(QColor(120,60,60));
        pen.setStyle(Qt::DotLine);
        selection->setPen(pen);
        selection->show();
        canvas()->update();
}

void FigureEditor::changeColor(const QColor & newColor)
{
        if (!selectedItem)
                return;
        if (selectedItem->rtti()==3)
                ((QCanvasText*)(selectedItem))->setColor(newColor);
        if (selectedItem->rtti()==5) {
                ((QCanvasRectangle*)(selectedItem))->setBrush(newColor);
                ((QCanvasRectangle*)(selectedItem))->setPen(QPen(newColor));
        }
        canvas()->update();
}


void FigureEditor::contentsMouseMoveEvent(QMouseEvent* e)
{
        QPoint p = inverseWorldMatrix().map(e->pos());

        // move item
        if ( moving ) {
                setCursor(QCursor(Qt::SizeAllCursor));
                selection->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
                moving->moveBy(p.x() - moving_start.x(), p.y() - moving_start.y());
                moving_start = p;
                if (operationMode!=CursorMode) {
                        operationMode=CursorMode;
                        setCursor(arrowCursor);
                        emit adjustButtons();
                }
        }
        // Creating rectangle
        else if (operationMode == RectMode || operationMode == ResizeMode) {
                if (drawingRect)
                        delete drawingRect;
                drawingRect = new QCanvasRectangle(QRect(draw_start,p),canvas());
                drawingRect->setPen(QPen(yellow));
                drawingRect->setZ(1001);
                drawingRect->show();
        }
        canvas()->setAllChanged ();
        canvas()->update();
}


void FigureEditor::exportContent()
{
        // Export content to a png image which can be used by mlt to create a video preview
        // All items are then drawed on the pixmap. To get transparency, it is required to
        // draw again all items on the alpha mask.

        QPixmap im(imageWidth,imageHeight);
        QPainter p;

        // Fill pixmap with color0, which sould be transparent but looks in fact to be black...
        im.fill(color0);

        // Create transparency mask
        im.setMask(im.createHeuristicMask());

        // Select all items
        QCanvasItemList list=canvas()->collisions(canvas()->rect());

        // Parse items in revers order to draw them on the pixmap
        QCanvasItemList::Iterator it = list.fromLast ();
        for (; it!=list.end(); --it) {
                if ( *it ) {
                        if ((*it)->rtti ()==3) // text item
                        {
                                p.begin(&im);
                                p.setPen(((QCanvasText*)(*it))->color());
                                p.setFont(((QCanvasText*)(*it))->font());
                                int wi=((QCanvasText*)(*it))->boundingRect().width()/2;
                                int he=((QCanvasText*)(*it))->boundingRect().height();
                                p.drawText(((QCanvasText*)(*it))->boundingRect(),Qt::AlignAuto,((QCanvasText*)(*it))->text());
                                p.end();

                                // Draw again on transparency mask
                                p.begin(im.mask());
                                p.setPen(((QCanvasText*)(*it))->color());
                                p.setFont(((QCanvasText*)(*it))->font());
                                p.drawText(((QCanvasText*)(*it))->boundingRect(),Qt::AlignAuto,((QCanvasText*)(*it))->text());
                                p.end();
                        }

                        if ((*it)->rtti ()==5 && (*it)->z()>=0 && (*it)->z()<1000) // rectangle item but don't draw the safe margins rectangle
                        {
                                p.begin(&im);
                                p.setPen(((QCanvasPolygonalItem*)(*it))->pen());
                                p.setBrush(((QCanvasPolygonalItem*)(*it))->brush());
                                p.drawRect((*it)->x(),(*it)->y(),((QCanvasRectangle*)(*it))->width(),((QCanvasRectangle*)(*it))->height());
                                p.end();

                                // Draw again on transparency mask
                                p.begin(im.mask());
                                p.setPen(QPen(color1,((QCanvasPolygonalItem*)(*it))->pen().width()));
                                p.setBrush(QBrush(color1));
                                p.drawRect((*it)->x(),(*it)->y(),((QCanvasRectangle*)(*it))->width(),((QCanvasRectangle*)(*it))->height());
                                p.end();
                        }
                }
        }

        // Save resulting pixmap in a file for mlt
        im.save(tmp->name(),"PNG");
        tmp->close();

        emit showPreview(tmp->name());
}


titleWidget::titleWidget( QWidget* parent, const char* name, WFlags fl ):
                titleBaseWidget(parent,name)
{
        canvas=new QCanvas(imageWidth,imageHeight);
        canview = new FigureEditor(*canvas,frame);

        // Put icons on buttons
        textButton->setPixmap(KGlobal::iconLoader()->loadIcon("title_text",KIcon::Small,22));
        rectButton->setPixmap(KGlobal::iconLoader()->loadIcon("title_rect",KIcon::Small,22));
        cursorButton->setPixmap(KGlobal::iconLoader()->loadIcon("arrow",KIcon::Toolbar));
        upButton->setPixmap(KGlobal::iconLoader()->loadIcon("up",KIcon::Toolbar));
        downButton->setPixmap(KGlobal::iconLoader()->loadIcon("down",KIcon::Toolbar));
        cursorButton->setOn(true);

        QHBoxLayout* flayout = new QHBoxLayout( frame, 1, 1, "flayout");
        flayout->addWidget( canview, 1 );

        QObject::connect(canview,SIGNAL(adjustButtons()),this,SLOT(adjustButtons()));
        QObject::connect(canview,SIGNAL(addText(QPoint)),this,SLOT(addText(QPoint)));
        QObject::connect(textButton,SIGNAL(clicked()),this,SLOT(textMode()));
        QObject::connect(cursorButton,SIGNAL(clicked()),this,SLOT(cursorMode()));
        QObject::connect(upButton,SIGNAL(clicked()),canview,SLOT(itemUp()));
        QObject::connect(downButton,SIGNAL(clicked()),canview,SLOT(itemDown()));
        QObject::connect(rectButton,SIGNAL(clicked()),this,SLOT(rectMode()));
        QObject::connect(canview,SIGNAL(addRect(QRect,int)),this,SLOT(addBlock(QRect,int)));
        QObject::connect(previewButton,SIGNAL(clicked()),this,SLOT(doPreview()));
        QObject::connect(fontSize,SIGNAL(valueChanged(int)),canview,SLOT(changeTextSize(int)));
        QObject::connect(fontFace,SIGNAL(textChanged(const QString &)),canview,SLOT(changeTextFace(const QString &)));
        QObject::connect(fontColor,SIGNAL(activated(const QColor &)),canview,SLOT(changeColor(const QColor &)));
        QObject::connect(canview,SIGNAL(editCanvasItem(QCanvasText*)),this,SLOT(editText(QCanvasText*)));
        QObject::connect(canview,SIGNAL(selectedCanvasItem(QCanvasText*)),this,SLOT(adjustWidgets(QCanvasText*)));
        QObject::connect(canview,SIGNAL(selectedCanvasItem(QCanvasRectangle*)),this,SLOT(adjustWidgets(QCanvasRectangle*)));
}


titleWidget::~titleWidget()
{}

void titleWidget::adjustButtons()
{
        // switch buttons on/off depending on the action
        cursorButton->setOn(canview->operationMode == CursorMode);
        textButton->setOn(canview->operationMode == TextMode);
        rectButton->setOn(canview->operationMode == RectMode);
}


void titleWidget::cursorMode()
{
        canview->setCursor(arrowCursor);
        canview->operationMode=CursorMode;
        adjustButtons();
}

void titleWidget::textMode()
{
        canview->setCursor(ibeamCursor);
        canview->operationMode=TextMode;
        adjustButtons();
}

void titleWidget::rectMode()
{
        canview->setCursor(crossCursor);
        canview->operationMode=RectMode;
        adjustButtons();
}

void titleWidget::addText(QPoint p)
{

        QString txt=KInputDialog::getMultiLineText("Enter your text",QString::null, QString::null,0,this);
        if (txt!=QString::null) {
                // Create new text item
                QCanvasText* i = new QCanvasText(canvas);
                i->setZ(canview->numItems);
                canview->numItems++;
                i->setText(txt);
                i->setFont(QFont(fontFace->currentFont(),fontSize->value()));
                i->setColor(fontColor->color());
                i->move(p.x(),p.y());
                i->show();

                // Select it
                delete canview->selection;
                canview->selection=0;
                canview->selectedItem=i;
                canview->selectRectangle(i);
                canvas->update();
        }
        canview->operationMode=CursorMode;
        canview->setCursor(arrowCursor);
        adjustButtons();

}


void titleWidget::addBlock(QRect rec,int pos)
{
        delete canview->selection;
        canview->selection=0;
        QCanvasRectangle* i = new QCanvasRectangle(rec,canvas);

        // Add new rectangle
        if (pos == -1) {
                i->setZ(canview->numItems);
                canview->numItems++;
        }
        // Resize existing rectangle
        else
                i->setZ(pos);

        i->setBrush(QBrush(fontColor->color()));
	QPen pen = QPen(fontColor->color());
        pen.setWidth(0);
        i->setPen(pen);
        i->show();

        // Select it
        canview->selectedItem=i;
        canview->selectRectangle(i);
        canview->operationMode=CursorMode;
        canview->setCursor(arrowCursor);
        canvas->update();
        adjustButtons();
}


void titleWidget::editText(QCanvasText* i)
{
        bool ok;
        QString txt=KInputDialog::getMultiLineText("Enter your text",QString::null, i->text(),&ok,this);
        if (!ok)
                return;
        // If empty text is returned, delete text item
        if (txt==QString::null) {
                canview->deleteItem((QCanvasItem *)(i));
                return;
        }
        i->setText(txt);
        i->setFont(QFont(fontFace->currentFont(),fontSize->value()));
        i->setColor(fontColor->color());
        i->show();
        canvas->update();
}


void titleWidget::adjustWidgets(QCanvasText* i)
{
        // Adjust font, size and color widgets according to the selected item
        fontFace->setCurrentFont(i->font().family());
        fontSize->setValue(i->font().pointSize ());
        fontColor->setColor(i->color());
}

void titleWidget::adjustWidgets(QCanvasRectangle* i)
{
        // Adjust color widget according to the selected item
        fontColor->setColor(i->brush().color());
}

void titleWidget::doPreview()
{
        // Prepare for mlt preview
        canview->exportContent();
}

